// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif
#include <iostream>
#include <cmath>
#include <dune/common/parallel/mpihelper.hh> // An initializer of MPI
#include <dune/common/exceptions.hh> // We use exceptions
#include <dune/grid/yaspgrid.hh>                // YetAnotherStructuredGrid
#include <dune/pdelab/finiteelementmap/qkfem.hh>
#include<dune/pdelab/backend/istl.hh>
#include <dune/pdelab/constraints/conforming.hh>
#include<dune/pdelab/function/callableadapter.hh>
#include <dune/geometry/referenceelements.hh>
#include<dune/grid/io/file/vtk.hh>
#include<dune/grid/io/file/vtk/vtkwriter.hh>
#include<dune/pdelab/common/vtkexport.hh>
#include<dune/pdelab/gridfunctionspace/vtk.hh>

int main(int argc, char** argv) {
		// Maybe initialize MPI
		Dune::MPIHelper& helper = Dune::MPIHelper::instance(argc, argv);
		if (Dune::MPIHelper::isFake) {
			std::cout << "SEQUENTIAL" << std::endl;
		} else {
			std::cout << "PARALLEL: rank " << helper.rank() << " of " << helper.size() << " processes." << std::endl;
		}
		const int dim = 2;
		//types:
		using GridType = Dune::YaspGrid<dim>;
		using GridView = typename GridType::LeafGridView;
		using FEM = Dune::PDELab::QkLocalFiniteElementMap<GridView, GridType::ctype,double,1>;
		using VBE0 = Dune::PDELab::ISTL::VectorBackend<>;
		using GFS = Dune::PDELab::GridFunctionSpace<GridView, FEM, VBE0>;
		using RHSType = Dune::PDELab::Backend::Vector<GFS,double>;
		using VTKWRITER =Dune::VTKWriter<GridView> ;
		using VTKSEQUENCEWRITER =Dune::VTKSequenceWriter<GridView> ;


		//build the grid and a pointers to it
		int overlap = 1;
		std::bitset < dim > B(false);
		Dune::FieldVector<double, dim> domain(1.0);
		std::array<int, dim> domainDims;
		domainDims[0] = 20;
		domainDims[1] = 20;
		std::shared_ptr<GridType> grid_p = std::shared_ptr < GridType> (new GridType(domain, domainDims, B, overlap, Dune::MPIHelper::getCollectiveCommunication()));
		std::shared_ptr<GridView> gv_p = std::shared_ptr < GridView > (std::make_shared<GridView>(grid_p->leafGridView()));

		//build FEM and gridfunctionspace
		std::shared_ptr<FEM> fem_p = std::shared_ptr < FEM > (new FEM(*gv_p));
		std::shared_ptr<GFS> gfs_p = std::shared_ptr < GFS > (new GFS(*gv_p, *fem_p));

		//build rhs vector z_p
		std::shared_ptr<RHSType> z_p = std::shared_ptr < RHSType > (new RHSType(*gfs_p));

		//interpolate dummy function into array: z_p
		const double pi = std::acos(-1);
		auto glambda = [&](const auto& global_coord) {return std::sin(global_coord[0]*(0.5*pi))*std::sin(global_coord[1]*(0.5*pi));};
		auto g = Dune::PDELab::makeGridFunctionFromCallable(*gv_p, glambda);
		Dune::PDELab::interpolate(g, *gfs_p, *z_p);

		//prepare output writer DGF's:

		//for field:
		using DiscreteGridFunction = Dune::PDELab::DiscreteGridFunction<GFS, RHSType>;
		using VTKF = Dune::PDELab::VTKGridFunctionAdapter<DiscreteGridFunction>;

		//For gradient-adapter:
		using DGFG =Dune::PDELab::DiscreteGridFunctionGradient<GFS,RHSType>;
		using VTKFG = Dune::PDELab::VTKGridFunctionAdapter<DGFG>;

		std::shared_ptr<DiscreteGridFunction> dgf_z_p = std::shared_ptr < DiscreteGridFunction > (new DiscreteGridFunction(*gfs_p, *z_p));

		VTKWRITER vtkwriter(*gv_p);
		std::string filename("output");
		VTKSEQUENCEWRITER vtkSequenceWriter(std::make_shared<VTKWRITER>(vtkwriter), filename, "./", "");

		//add the two fields
		vtkSequenceWriter.addVertexData(std::shared_ptr < VTKF > (new VTKF(*dgf_z_p, "field")));
		vtkSequenceWriter.addVertexData(std::make_shared<VTKFG>(std::make_shared<DGFG>(*gfs_p, *z_p), "gradient"));

		std::cout << "writing output: " << filename << std::endl;
		vtkSequenceWriter.write(0, Dune::VTK::appendedraw);

		return 0;
}
